# ruff: noqa: UP007, UP006
from __future__ import annotations

import functools
import inspect
from enum import Enum
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import strawberry
from django.db.models import Q, QuerySet
from strawberry import UNSET, relay
from strawberry.field import StrawberryField, field
from strawberry.type import WithStrawberryObjectDefinition, has_object_definition
from strawberry.unset import UnsetType
from typing_extensions import Self, assert_never, dataclass_transform

from strawberry_django.fields.filter_order import (
    RESOLVE_VALUE_META,
    WITH_NONE_META,
    FilterOrderField,
    FilterOrderFieldResolver,
)
from strawberry_django.utils.typing import (
    WithStrawberryDjangoObjectDefinition,
    has_django_definition,
)

from .arguments import argument
from .fields.base import StrawberryDjangoFieldBase
from .settings import strawberry_django_settings

if TYPE_CHECKING:
    from types import FunctionType

    from django.db.models import Model
    from strawberry.arguments import StrawberryArgument
    from strawberry.types import Info


T = TypeVar("T")
_T = TypeVar("_T", bound=type)
_QS = TypeVar("_QS", bound="QuerySet")

FILTERS_ARG = "filters"


@strawberry.input
class DjangoModelFilterInput:
    pk: strawberry.ID


@strawberry.input
class FilterLookup(Generic[T]):
    exact: Optional[T] = UNSET
    i_exact: Optional[T] = UNSET
    contains: Optional[T] = UNSET
    i_contains: Optional[T] = UNSET
    in_list: Optional[List[T]] = UNSET
    gt: Optional[T] = UNSET
    gte: Optional[T] = UNSET
    lt: Optional[T] = UNSET
    lte: Optional[T] = UNSET
    starts_with: Optional[T] = UNSET
    i_starts_with: Optional[T] = UNSET
    ends_with: Optional[T] = UNSET
    i_ends_with: Optional[T] = UNSET
    range: Optional[List[T]] = UNSET
    is_null: Optional[bool] = UNSET
    regex: Optional[str] = UNSET
    i_regex: Optional[str] = UNSET


lookup_name_conversion_map = {
    "i_exact": "iexact",
    "i_contains": "icontains",
    "in_list": "in",
    "starts_with": "startswith",
    "i_starts_with": "istartswith",
    "ends_with": "endswith",
    "i_ends_with": "iendswith",
    "is_null": "isnull",
    "i_regex": "iregex",
}


def resolve_value(value: Any) -> Any:
    if isinstance(value, list):
        return [resolve_value(v) for v in value]

    if isinstance(value, relay.GlobalID):
        return value.node_id

    if isinstance(value, Enum):
        return value.value

    return value


@functools.lru_cache(maxsize=256)
def _function_allow_passing_info(filter_method: FunctionType) -> bool:
    argspec = inspect.getfullargspec(filter_method)

    return "info" in getattr(argspec, "args", []) or "info" in getattr(
        argspec,
        "kwargs",
        [],
    )


def _process_deprecated_filter(
    filter_method: FunctionType, info: Info | None, queryset: _QS
) -> _QS:
    kwargs = {}
    if _function_allow_passing_info(
        # Pass the original __func__ which is always the same
        getattr(filter_method, "__func__", filter_method),
    ):
        kwargs["info"] = info

    return filter_method(queryset=queryset, **kwargs)


def process_filters(
    filters: WithStrawberryObjectDefinition,
    queryset: _QS,
    info: Info | None,
    prefix: str = "",
    skip_object_filter_method: bool = False,
) -> Tuple[_QS, Q]:
    using_old_filters = strawberry_django_settings()["USE_DEPRECATED_FILTERS"]

    q = Q()

    if not skip_object_filter_method and (
        filter_method := getattr(filters, "filter", None)
    ):
        # Dedicated function for object
        if isinstance(filter_method, FilterOrderFieldResolver):
            return filter_method(filters, info, queryset=queryset, prefix=prefix)
        if using_old_filters:
            return _process_deprecated_filter(filter_method, info, queryset), q

    # This loop relies on the filter field order that is not quaranteed for GQL input objects:
    #   "filter" has to be first since it overrides filtering for entire object
    #   DISTINCT has to be last and OR has to be after because it must be
    #       applied agains all other since default connector is AND
    for f in sorted(
        filters.__strawberry_definition__.fields,
        key=lambda x: len(x.name) if x.name in {"OR", "DISTINCT"} else 0,
    ):
        field_value = getattr(filters, f.name)
        # None is still acceptable for v1 (backwards compatibility) and filters that support it via metadata
        if field_value is UNSET or (
            field_value is None
            and not f.metadata.get(WITH_NONE_META, using_old_filters)
        ):
            continue

        should_resolve = f.metadata.get(RESOLVE_VALUE_META, UNSET)

        field_name = lookup_name_conversion_map.get(f.name, f.name)
        if field_name == "DISTINCT":
            if field_value:
                queryset = queryset.distinct()
        elif field_name in ("AND", "OR", "NOT"):  # noqa: PLR6201
            assert has_object_definition(field_value)

            queryset, sub_q = process_filters(
                cast(WithStrawberryObjectDefinition, field_value),
                queryset,
                info,
                prefix,
            )
            if field_name == "AND":
                q &= sub_q
            elif field_name == "OR":
                q |= sub_q
            elif field_name == "NOT":
                q &= ~sub_q
            else:
                assert_never(field_name)
        elif isinstance(f, FilterOrderField) and f.base_resolver:
            res = f.base_resolver(
                filters,
                info,
                value=(resolve_value(field_value) if should_resolve else field_value),
                queryset=queryset,
                prefix=prefix,
            )
            if isinstance(res, tuple):
                queryset, sub_q = res
            else:
                sub_q = res

            q &= sub_q
        elif using_old_filters and (
            filter_method := getattr(filters, f"filter_{field_name}", None)
        ):
            queryset = _process_deprecated_filter(filter_method, info, queryset)
        elif has_object_definition(field_value):
            queryset, sub_q = process_filters(
                cast(WithStrawberryObjectDefinition, field_value),
                queryset,
                info,
                f"{prefix}{field_name}__",
            )
            q &= sub_q
        else:
            q &= Q(**{
                f"{prefix}{field_name}": (
                    resolve_value(field_value)
                    if should_resolve or should_resolve is UNSET
                    else field_value
                )
            })

    return queryset, q


def apply(
    filters: object | None,
    queryset: _QS,
    info: Info | None = None,
    pk: Any | None = None,
) -> _QS:
    if pk not in (None, strawberry.UNSET):  # noqa: PLR6201
        queryset = queryset.filter(pk=pk)

    if filters in (None, strawberry.UNSET) or not has_django_definition(filters):  # noqa: PLR6201
        return queryset

    queryset, q = process_filters(
        cast(WithStrawberryObjectDefinition, filters), queryset, info
    )
    if q:
        queryset = queryset.filter(q)
    return queryset


class StrawberryDjangoFieldFilters(StrawberryDjangoFieldBase):
    def __init__(self, filters: Union[type, UnsetType, None] = UNSET, **kwargs):
        if filters and not has_object_definition(filters):
            raise TypeError("filters needs to be a strawberry type")

        self.filters = filters
        super().__init__(**kwargs)

    def __copy__(self) -> Self:
        new_field = super().__copy__()
        new_field.filters = self.filters
        return new_field

    @property
    def arguments(self) -> List[StrawberryArgument]:
        arguments = []
        if self.base_resolver is None:
            filters = self.get_filters()
            origin = cast(WithStrawberryObjectDefinition, self.origin)
            is_root_query = origin.__strawberry_definition__.name == "Query"

            if (
                self.django_model
                and is_root_query
                and isinstance(self.django_type, relay.Node)
            ):
                arguments.append(
                    (
                        argument("ids", List[relay.GlobalID])
                        if self.is_list
                        else argument("id", relay.GlobalID)
                    ),
                )
            if (
                self.django_model
                and is_root_query
                and not self.is_list
                and not self.is_connection
            ):
                arguments.append(argument("pk", strawberry.ID))
            elif filters is not None and self.is_list:
                arguments.append(argument(FILTERS_ARG, filters, is_optional=True))

        return super().arguments + arguments

    @arguments.setter
    def arguments(self, value: list[StrawberryArgument]):
        args_prop = super(StrawberryDjangoFieldFilters, self.__class__).arguments
        return args_prop.fset(self, value)  # type: ignore

    def get_filters(self) -> type[WithStrawberryObjectDefinition] | None:
        filters = self.filters
        if filters is None:
            return None

        if isinstance(filters, UnsetType):
            django_type = self.django_type
            filters = (
                django_type.__strawberry_django_definition__.filters
                if django_type is not None
                else None
            )

        return filters if filters is not UNSET else None

    def get_queryset(
        self,
        queryset: _QS,
        info: Info,
        *,
        filters: Optional[WithStrawberryDjangoObjectDefinition] = None,
        pk: Optional[Any] = None,
        **kwargs,
    ) -> _QS:
        queryset = super().get_queryset(queryset, info, **kwargs)
        return apply(filters, queryset, info, pk)


@dataclass_transform(
    order_default=True,
    field_specifiers=(
        StrawberryField,
        field,
    ),
)
def filter(  # noqa: A001
    model: type[Model],
    *,
    name: str | None = None,
    description: str | None = None,
    directives: Sequence[object] | None = (),
    lookups: bool = False,
) -> Callable[[_T], _T]:
    from .type import input

    return input(
        model,
        name=name,
        description=description,
        directives=directives,
        is_filter="lookups" if lookups else True,
        partial=True,
    )



import datetime
import decimal
import uuid
from typing import (
    Generic,
    List,
    Optional,
    TypeVar,
)

import strawberry
from django.db.models import Q
from strawberry import UNSET

from strawberry_django.filters import resolve_value

from .filter_order import filter_field

T = TypeVar("T")

_SKIP_MSG = "Filter will be skipped on `null` value"


@strawberry.input
class BaseFilterLookup(Generic[T]):
    exact: Optional[T] = filter_field(description=f"Exact match. {_SKIP_MSG}")
    is_null: Optional[bool] = filter_field(description=f"Assignment test. {_SKIP_MSG}")
    in_list: Optional[List[T]] = filter_field(
        description=f"Exact match of items in a given list. {_SKIP_MSG}"
    )


@strawberry.input
class RangeLookup(Generic[T]):
    start: Optional[T] = None
    end: Optional[T] = None

    @filter_field
    def filter(self, queryset, prefix: str):
        return queryset, Q(**{
            prefix[:-2]: [resolve_value(self.start), resolve_value(self.end)]
        })


@strawberry.input
class ComparisonFilterLookup(BaseFilterLookup[T]):
    gt: Optional[T] = filter_field(description=f"Greater than. {_SKIP_MSG}")
    gte: Optional[T] = filter_field(
        description=f"Greater than or equal to. {_SKIP_MSG}"
    )
    lt: Optional[T] = filter_field(description=f"Less than. {_SKIP_MSG}")
    lte: Optional[T] = filter_field(description=f"Less than or equal to. {_SKIP_MSG}")
    range: Optional[RangeLookup[T]] = filter_field(
        description="Inclusive range test (between)"
    )


@strawberry.input
class FilterLookup(BaseFilterLookup[T]):
    i_exact: Optional[T] = filter_field(
        description=f"Case-insensitive exact match. {_SKIP_MSG}"
    )
    contains: Optional[T] = filter_field(
        description=f"Case-sensitive containment test. {_SKIP_MSG}"
    )
    i_contains: Optional[T] = filter_field(
        description=f"Case-insensitive containment test. {_SKIP_MSG}"
    )
    starts_with: Optional[T] = filter_field(
        description=f"Case-sensitive starts-with. {_SKIP_MSG}"
    )
    i_starts_with: Optional[T] = filter_field(
        description=f"Case-insensitive starts-with. {_SKIP_MSG}"
    )
    ends_with: Optional[T] = filter_field(
        description=f"Case-sensitive ends-with. {_SKIP_MSG}"
    )
    i_ends_with: Optional[T] = filter_field(
        description=f"Case-insensitive ends-with. {_SKIP_MSG}"
    )
    regex: Optional[T] = filter_field(
        description=f"Case-sensitive regular expression match. {_SKIP_MSG}"
    )
    i_regex: Optional[T] = filter_field(
        description=f"Case-insensitive regular expression match. {_SKIP_MSG}"
    )


@strawberry.input
class DateFilterLookup(ComparisonFilterLookup[T]):
    year: Optional[ComparisonFilterLookup[int]] = UNSET
    month: Optional[ComparisonFilterLookup[int]] = UNSET
    day: Optional[ComparisonFilterLookup[int]] = UNSET
    week_day: Optional[ComparisonFilterLookup[int]] = UNSET
    iso_week_day: Optional[ComparisonFilterLookup[int]] = UNSET
    week: Optional[ComparisonFilterLookup[int]] = UNSET
    iso_year: Optional[ComparisonFilterLookup[int]] = UNSET
    quarter: Optional[ComparisonFilterLookup[int]] = UNSET


@strawberry.input
class TimeFilterLookup(ComparisonFilterLookup[T]):
    hour: Optional[ComparisonFilterLookup[int]] = UNSET
    minute: Optional[ComparisonFilterLookup[int]] = UNSET
    second: Optional[ComparisonFilterLookup[int]] = UNSET
    date: Optional[ComparisonFilterLookup[int]] = UNSET
    time: Optional[ComparisonFilterLookup[int]] = UNSET


@strawberry.input
class DatetimeFilterLookup(DateFilterLookup[T], TimeFilterLookup[T]):
    pass


type_filter_map = {
    strawberry.ID: BaseFilterLookup,
    bool: BaseFilterLookup,
    datetime.date: DateFilterLookup,
    datetime.datetime: DatetimeFilterLookup,
    datetime.time: TimeFilterLookup,
    decimal.Decimal: ComparisonFilterLookup,
    float: ComparisonFilterLookup,
    int: ComparisonFilterLookup,
    str: FilterLookup,
    uuid.UUID: FilterLookup,
}



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, Type, cast

import strawberry
from asgiref.sync import async_to_sync
from django.contrib import auth
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import ValidationError

from strawberry_django.auth.utils import get_current_user
from strawberry_django.mutations import mutations, resolvers
from strawberry_django.mutations.fields import DjangoCreateMutation
from strawberry_django.optimizer import DjangoOptimizerExtension
from strawberry_django.resolvers import django_resolver
from strawberry_django.utils.requests import get_request

try:
    # Django-channels is not always used/intalled,
    # therefore it shouldn't be it a hard requirement.
    from channels import auth as channels_auth
except ModuleNotFoundError:
    channels_auth = None

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser
    from strawberry.types import Info


@django_resolver
def resolve_login(info: Info, username: str, password: str) -> AbstractBaseUser:
    request = get_request(info)
    user = auth.authenticate(request, username=username, password=password)

    if user is None:
        raise ValidationError("Incorrect username/password")

    try:
        auth.login(request, user)
    except AttributeError:
        # ASGI in combo with websockets needs the channels login functionality.
        # to ensure we're talking about channels, let's veriy that our
        # request is actually channelsrequest
        try:
            scope = request.consumer.scope  # type: ignore
            async_to_sync(channels_auth.login)(scope, user)  # type: ignore
            # According to channels docs you must save the session
            scope["session"].save()
        except (AttributeError, NameError):
            # When Django-channels is not installed,
            # this code will be non-existing
            pass

    return user


@django_resolver
def resolve_logout(info: Info) -> bool:
    user = get_current_user(info)
    ret = user.is_authenticated

    try:
        request = get_request(info)
        auth.logout(request)
    except AttributeError:
        try:
            scope = request.consumer.scope  # type: ignore
            async_to_sync(channels_auth.logout)(scope)  # type: ignore
        except (AttributeError, NameError):
            # When Django-channels is not installed,
            # this code will be non-existing
            pass

    return ret


class DjangoRegisterMutation(DjangoCreateMutation):
    def create(self, data: dict[str, Any], *, info: Info):
        model = cast(Type["AbstractBaseUser"], self.django_model)
        assert model is not None

        password = data.pop("password")
        validate_password(password)

        # Do not optimize anything while retrieving the object to update
        with DjangoOptimizerExtension.disabled():
            return resolvers.create(
                info,
                model,
                data,
                key_attr=self.key_attr,
                full_clean=self.full_clean,
                pre_save_hook=lambda obj: obj.set_password(password),
            )


login = functools.partial(strawberry.mutation, resolver=resolve_login)
logout = functools.partial(strawberry.mutation, resolver=resolve_logout)
register = mutations.create if TYPE_CHECKING else DjangoRegisterMutation



# Based on https://github.com/flavors/django-graphiql-debug-toolbar

import asyncio
import collections
import contextlib
import json
import weakref
from typing import Optional

from asgiref.sync import iscoroutinefunction, markcoroutinefunction, sync_to_async
from debug_toolbar.middleware import _HTML_TYPES, get_show_toolbar  # noqa: PLC2701
from debug_toolbar.middleware import (
    DebugToolbarMiddleware as _DebugToolbarMiddleware,
)
from debug_toolbar.panels.sql.panel import SQLPanel
from debug_toolbar.panels.templates import TemplatesPanel
from debug_toolbar.toolbar import DebugToolbar
from django.core.exceptions import SynchronousOnlyOperation
from django.core.serializers.json import DjangoJSONEncoder
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.template.loader import render_to_string
from django.utils.encoding import force_str
from strawberry.django.views import BaseView

_store_cache = weakref.WeakKeyDictionary()
_debug_toolbar_map: "weakref.WeakKeyDictionary[HttpRequest, DebugToolbar]" = (
    weakref.WeakKeyDictionary()
)

_original_store = DebugToolbar.store
_original_debug_toolbar_init = DebugToolbar.__init__
_original_store_template_info = TemplatesPanel._store_template_info


def _is_websocket(request: HttpRequest):
    return (
        request.META.get("HTTP_UPGRADE") == "websocket"
        and request.META.get("HTTP_CONNECTION") == "Upgrade"
    )


def _debug_toolbar_init(self, request, *args, **kwargs):
    _debug_toolbar_map[request] = self
    _original_debug_toolbar_init(self, request, *args, **kwargs)
    self.config["RENDER_PANELS"] = False
    self.config["SKIP_TEMPLATE_PREFIXES"] = (
        *tuple(self.config.get("SKIP_TEMPLATE_PREFIXES", [])),
        "graphql/",
    )


def _store(toolbar: DebugToolbar):
    _debug_toolbar_map[toolbar.request] = toolbar
    _original_store(toolbar)
    _store_cache[toolbar.request] = toolbar.store_id


def _store_template_info(*args, **kwargs):
    with contextlib.suppress(SynchronousOnlyOperation):
        return _original_store_template_info(*args, **kwargs)


def _get_payload(request: HttpRequest, response: HttpResponse):
    store_id = _store_cache.get(request)
    if not store_id:
        return None

    toolbar: Optional[DebugToolbar] = DebugToolbar.fetch(store_id)
    if not toolbar:
        return None

    content = force_str(response.content, encoding=response.charset)
    payload = json.loads(content, object_pairs_hook=collections.OrderedDict)
    payload["debugToolbar"] = collections.OrderedDict(
        [("panels", collections.OrderedDict())],
    )
    payload["debugToolbar"]["storeId"] = toolbar.store_id

    for p in reversed(toolbar.enabled_panels):
        if p.panel_id == "TemplatesPanel":
            continue

        title = p.title if p.has_content else None

        sub = p.nav_subtitle
        payload["debugToolbar"]["panels"][p.panel_id] = {
            "title": title() if callable(title) else title,
            "subtitle": sub() if callable(sub) else sub,
        }

    return payload


DebugToolbar.__init__ = _debug_toolbar_init
DebugToolbar.store = _store  # type: ignore
TemplatesPanel._store_template_info = _store_template_info


class DebugToolbarMiddleware(_DebugToolbarMiddleware):
    sync_capable = True
    async_capable = True

    def __init__(self, get_response):
        self._original_get_response = get_response

        if iscoroutinefunction(get_response):
            markcoroutinefunction(self)

            def _get_response(request):
                toolbar = _debug_toolbar_map.pop(request, None)
                for panel in toolbar.enabled_panels if toolbar else []:
                    if isinstance(panel, SQLPanel):
                        sql_panel = panel
                        break
                else:
                    sql_panel = None

                async def _inner_get_response():
                    if sql_panel:
                        await sync_to_async(sql_panel.enable_instrumentation)()
                    try:
                        return await self._original_get_response(request)
                    finally:
                        if sql_panel:
                            await sync_to_async(sql_panel.disable_instrumentation)()

                return asyncio.run(_inner_get_response())

            get_response = _get_response

        super().__init__(get_response)

    def __call__(self, request: HttpRequest):
        if iscoroutinefunction(self):
            return self.__acall__(request)

        if _is_websocket(request):
            return self._original_get_response(request)

        return self.process_request(request)

    async def __acall__(self, request: HttpRequest):  # noqa: PLW3201
        if _is_websocket(request):
            return await self._original_get_response(request)

        return await sync_to_async(self.process_request, thread_sensitive=False)(
            request,
        )

    def process_request(self, request: HttpRequest):
        response = super().__call__(request)

        show_toolbar = get_show_toolbar()
        if (
            callable(show_toolbar) and not show_toolbar(request)
        ) or DebugToolbar.is_toolbar_request(request):
            return response

        content_type = response.get("Content-Type", "").split(";")[0]
        is_html = content_type in _HTML_TYPES
        is_graphiql = getattr(request, "_is_graphiql", False)

        if is_html and is_graphiql and response.status_code == 200:  # noqa: PLR2004
            template = render_to_string("strawberry_django/debug_toolbar.html")
            response.write(template)
            if "Content-Length" in response:
                response["Content-Length"] = len(response.content)

        if is_html or not is_graphiql or content_type != "application/json":
            return response

        try:
            operation_name = json.loads(request.body).get("operationName")
        except Exception:  # noqa: BLE001
            operation_name = None

        # Do not return the payload for introspection queries, otherwise IDEs such as
        # apollo sandbox that query the introspection all the time will remove older
        # results from the history.
        payload = (
            _get_payload(request, response)
            if operation_name != "IntrospectionQuery"
            else None
        )
        if payload is None:
            return response

        response.content = json.dumps(payload, cls=DjangoJSONEncoder)
        if "Content-Length" in response:
            response["Content-Length"] = len(response.content)

        return response

    def process_view(self, request: HttpRequest, view_func, *args, **kwargs):
        view = getattr(view_func, "view_class", None)
        request._is_graphiql = bool(view and issubclass(view, BaseView))  # type: ignore



from __future__ import annotations

import contextlib
import contextvars
import copy
import dataclasses
import itertools
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ForwardRef,
    Generator,
    Type,
    TypeVar,
    cast,
)

from django.db import models
from django.db.models import Prefetch
from django.db.models.constants import LOOKUP_SEP
from django.db.models.expressions import BaseExpression, Combinable
from django.db.models.fields.reverse_related import (
    ManyToManyRel,
    ManyToOneRel,
    OneToOneRel,
)
from django.db.models.manager import BaseManager
from django.db.models.query import QuerySet
from graphql.language.ast import OperationType
from graphql.type.definition import GraphQLResolveInfo, get_named_type
from strawberry import relay
from strawberry.extensions import SchemaExtension
from strawberry.lazy_type import LazyType
from strawberry.object_type import StrawberryObjectDefinition
from strawberry.schema.schema import Schema
from strawberry.type import get_object_definition
from strawberry.types.info import Info
from strawberry.types.nodes import InlineFragment, Selection, convert_selections
from strawberry.utils.typing import eval_type
from typing_extensions import assert_never, assert_type, get_args

from strawberry_django.fields.types import resolve_model_field_name
from strawberry_django.resolvers import django_fetch

from .descriptors import ModelProperty
from .utils.inspect import (
    PrefetchInspector,
    get_model_fields,
    get_possible_type_definitions,
    get_selections,
)
from .utils.typing import (
    AnnotateCallable,
    AnnotateType,
    PrefetchCallable,
    PrefetchType,
    TypeOrMapping,
    TypeOrSequence,
    get_django_definition,
)

if TYPE_CHECKING:
    from strawberry.types.execution import ExecutionContext
    from strawberry.utils.await_maybe import AwaitableOrValue


__all__ = [
    "DjangoOptimizerExtension",
    "OptimizerConfig",
    "OptimizerStore",
    "PrefetchType",
    "optimize",
]

_M = TypeVar("_M", bound=models.Model)

_sentinel = object()
_annotate_placeholder = "__annotated_placeholder__"
_interfaces: defaultdict[
    Schema,
    dict[StrawberryObjectDefinition, list[StrawberryObjectDefinition]],
] = defaultdict(
    dict,
)


@dataclasses.dataclass
class OptimizerConfig:
    """Django optimization configuration.

    Attributes
    ----------
        enable_only:
            Enable `QuerySet.only` optimizations
        enable_select_related:
            Enable `QuerySet.select_related` optimizations
        enable_prefetch_related:
            Enable `QuerySet.prefetch_related` optimizations
        enable_annotate:
            Enable `QuerySet.annotate` optimizations
        prefetch_custom_queryset:
            Use custom instead of _base_manager for prefetch querysets

    """

    enable_only: bool = dataclasses.field(default=True)
    enable_select_related: bool = dataclasses.field(default=True)
    enable_prefetch_related: bool = dataclasses.field(default=True)
    enable_annotate: bool = dataclasses.field(default=True)
    prefetch_custom_queryset: bool = dataclasses.field(default=False)


@dataclasses.dataclass
class OptimizerStore:
    """Django optimization store.

    Attributes
    ----------
        only:
            Set of values to optimize using `QuerySet.only`
        selected:
            Set of values to optimize using `QuerySet.select_related`
        prefetch_related:
            Set of values to optimize using `QuerySet.prefetch_related`
        annotate:
            Dict of values to use in `QuerySet.annotate`

    """

    only: list[str] = dataclasses.field(default_factory=list)
    select_related: list[str] = dataclasses.field(default_factory=list)
    prefetch_related: list[PrefetchType] = dataclasses.field(default_factory=list)
    annotate: dict[str, AnnotateType] = dataclasses.field(default_factory=dict)

    def __bool__(self):
        return any(
            [self.only, self.select_related, self.prefetch_related, self.annotate],
        )

    def __ior__(self, other: OptimizerStore):
        self.only.extend(other.only)
        self.select_related.extend(other.select_related)
        self.prefetch_related.extend(other.prefetch_related)
        self.annotate.update(other.annotate)
        return self

    def __or__(self, other: OptimizerStore):
        new = self.copy()
        new |= other
        return new

    def copy(self):
        return self.__class__(
            only=self.only[:],
            select_related=self.select_related[:],
            prefetch_related=self.prefetch_related[:],
            annotate=self.annotate.copy(),
        )

    @classmethod
    def with_hints(
        cls,
        *,
        only: TypeOrSequence[str] | None = None,
        select_related: TypeOrSequence[str] | None = None,
        prefetch_related: TypeOrSequence[PrefetchType] | None = None,
        annotate: TypeOrMapping[AnnotateType] | None = None,
    ):
        return cls(
            only=[only] if isinstance(only, str) else list(only or []),
            select_related=(
                [select_related]
                if isinstance(select_related, str)
                else list(select_related or [])
            ),
            prefetch_related=(
                [prefetch_related]
                if isinstance(prefetch_related, (str, Prefetch, Callable))
                else list(prefetch_related or [])
            ),
            annotate=(
                # placeholder here,
                # because field name is evaluated later on .annotate call:
                {_annotate_placeholder: annotate}
                if isinstance(annotate, (BaseExpression, Combinable, Callable))
                else dict(annotate or {})
            ),
        )

    def with_prefix(self, prefix: str, *, info: GraphQLResolveInfo):
        prefetch_related = []
        for p in self.prefetch_related:
            if isinstance(p, Callable):
                assert_type(p, PrefetchCallable)
                p = p(info)  # noqa: PLW2901

            if isinstance(p, str):
                prefetch_related.append(f"{prefix}{LOOKUP_SEP}{p}")
            elif isinstance(p, Prefetch):
                # add_prefix modifies the field's prefetch object, so we copy it before
                p_copy = copy.copy(p)
                p_copy.add_prefix(prefix)
                prefetch_related.append(p_copy)
            else:  # pragma:nocover
                assert_never(p)

        annotate = {}
        for k, v in self.annotate.items():
            if isinstance(v, Callable):
                assert_type(v, AnnotateCallable)
                v = v(info)  # noqa: PLW2901
            annotate[f"{prefix}{LOOKUP_SEP}{k}"] = v

        return self.__class__(
            only=[f"{prefix}{LOOKUP_SEP}{i}" for i in self.only],
            select_related=[f"{prefix}{LOOKUP_SEP}{i}" for i in self.select_related],
            prefetch_related=prefetch_related,
            annotate=annotate,
        )

    def apply(
        self,
        qs: QuerySet[_M],
        *,
        info: GraphQLResolveInfo,
        config: OptimizerConfig | None = None,
    ) -> QuerySet[_M]:
        config = config or OptimizerConfig()

        if config.enable_prefetch_related and self.prefetch_related:
            # Add all str at the same time to make it easier to handle Prefetch below
            to_prefetch: dict[str, str | Prefetch] = {
                p: p for p in self.prefetch_related if isinstance(p, str)
            }

            abort_only = set()
            # Merge already existing prefetches together
            for p in itertools.chain(
                qs._prefetch_related_lookups,  # type: ignore
                self.prefetch_related,
            ):
                # Already added above
                if isinstance(p, str):
                    continue

                if isinstance(p, Callable):
                    assert_type(p, PrefetchCallable)
                    p = p(info)  # noqa: PLW2901

                path = cast(str, p.prefetch_to)  # type: ignore
                existing = to_prefetch.get(path)
                # The simplest case. The prefetch doesn't exist or is a string.
                # In this case, just replace it.
                if not existing or isinstance(existing, str):
                    to_prefetch[path] = p
                    if isinstance(existing, str):
                        abort_only.add(path)
                    continue

                p1 = PrefetchInspector(existing)
                p2 = PrefetchInspector(p)
                if getattr(existing, "_optimizer_sentinel", None) is _sentinel:
                    ret = p1.merge(p2, allow_unsafe_ops=True)
                elif getattr(p, "_optimizer_sentinel", None) is _sentinel:
                    ret = p2.merge(p1, allow_unsafe_ops=True)
                else:
                    # The order here doesn't matter
                    ret = p1.merge(p2)

                to_prefetch[path] = ret.prefetch

            # Abort only optimization if one prefetch related was made for everything
            for ao in abort_only:
                to_prefetch[ao].queryset.query.deferred_loading = (  # type: ignore
                    [],
                    True,
                )

            # First prefetch_related(None) to clear all existing prefetches, and them
            # add ours, which also contains them. This is to avoid the
            # "lookup was already seen with a different queryset" error
            qs = qs.prefetch_related(None).prefetch_related(*to_prefetch.values())

        only_set = set(self.only)
        select_related_only_set = set()

        if config.enable_select_related and self.select_related:
            qs = qs.select_related(*self.select_related)

            for select_related in self.select_related:
                if select_related in only_set:
                    continue

                if not any(only.startswith(select_related) for only in only_set):
                    select_related_only_set.add(select_related)

        if config.enable_only and only_set:
            qs = qs.only(*(only_set | select_related_only_set))

        if config.enable_annotate and self.annotate:
            to_annotate = {}
            for k, v in self.annotate.items():
                if isinstance(v, Callable):
                    assert_type(v, AnnotateCallable)
                    v = v(info)  # noqa: PLW2901
                to_annotate[k] = v
            qs = qs.annotate(**to_annotate)

        return qs


def _get_prefetch_queryset(
    remote_model: type[models.Model],
    field,
    config: OptimizerConfig | None,
    info: GraphQLResolveInfo,
) -> QuerySet:
    qs = remote_model._base_manager.all()  # type: ignore
    if not config or not config.prefetch_custom_queryset:
        return qs

    remote_type_defs = get_args(field.type_annotation.annotation)
    if len(remote_type_defs) != 1:
        raise TypeError(f"Expected exactly one remote type: {remote_type_defs}")
    if type(remote_type_defs[0]) is ForwardRef:
        remote_type = eval_type(
            remote_type_defs[0],
            field.type_annotation.namespace,
            None,
        )
    else:
        remote_type = remote_type_defs[0]

    if get_queryset := getattr(remote_type, "get_queryset", None):
        return get_queryset(qs, info)

    return qs


def _get_model_hints(
    model: type[models.Model],
    schema: Schema,
    object_definition: StrawberryObjectDefinition,
    selection: Selection,
    *,
    info: GraphQLResolveInfo,
    config: OptimizerConfig | None = None,
    prefix: str = "",
    cache: dict[type[models.Model], list[tuple[int, OptimizerStore]]] | None = None,
    level: int = 0,
) -> OptimizerStore | None:
    try:
        from django.contrib.contenttypes.fields import (
            GenericForeignKey,
            GenericRelation,
        )
    except (ImportError, RuntimeError):  # pragma: no cover
        GenericForeignKey = None  # noqa: N806
        GenericRelation = None  # noqa: N806
        _relation_fields = (models.ManyToManyField, ManyToManyRel, ManyToOneRel)
    else:
        _relation_fields = (
            models.ManyToManyField,
            ManyToManyRel,
            ManyToOneRel,
            GenericRelation,
        )

    store = OptimizerStore()
    cache = cache or {}
    t_name = schema.config.name_converter.from_object(object_definition)

    # In case this is a relay field, find the selected edges/nodes, the selected fields
    # are actually inside edges -> node selection...
    if issubclass(
        object_definition.origin,
        relay.Connection,
    ):
        # TODO: Connections are mostly used for pagination so it doesn't make sense for
        # us to optimize those, as our prefetch would be thrown away causing an extra
        # useless query. Is there a way for us to properly optimize this in the future?
        if level > 0:
            return None

        n_type = object_definition.type_var_map.get("NodeType")
        if n_type is None:
            specialized_type_var_map = object_definition.specialized_type_var_map or {}

            n_type = specialized_type_var_map["NodeType"]
        if isinstance(n_type, LazyType):
            n_type = n_type.resolve_type()

        n_definition = get_object_definition(n_type, strict=True)

        for edges in get_selections(selection, typename=t_name).values():
            if edges.name != "edges":
                continue

            e_definition = get_object_definition(relay.Edge, strict=True)
            e_type = e_definition.resolve_generic(
                relay.Edge[cast(Type[relay.Node], n_type)],
            )
            e_name = schema.config.name_converter.from_object(
                get_object_definition(e_type, strict=True),
            )
            for node in get_selections(edges, typename=e_name).values():
                if node.name != "node":
                    continue

                new_store = _get_model_hints(
                    model=model,
                    schema=schema,
                    object_definition=n_definition,
                    selection=node,
                    info=info,
                    config=config,
                    prefix=prefix,
                    cache=cache,
                    level=level,
                )
                if new_store is not None:
                    store |= new_store

        return store

    fields = {
        schema.config.name_converter.get_graphql_name(f): f
        for f in object_definition.fields
    }
    model_fields = get_model_fields(model)

    dj_definition = get_django_definition(object_definition.origin)
    if (
        dj_definition is None
        or not issubclass(model, dj_definition.model)
        or dj_definition.disable_optimization
    ):
        return None

    dj_type_store = getattr(dj_definition, "store", None)
    if dj_type_store:
        store |= dj_type_store

    # Make sure that the model's pk is always selected when using only
    pk = model._meta.pk
    if pk is not None:
        store.only.append(pk.attname)

    for f_selection in get_selections(selection, typename=t_name).values():
        field = fields.get(f_selection.name, None)
        if not field:
            continue

        # Do not optimize the field if the user asked not to
        if getattr(field, "disable_optimization", False):
            continue

        # Add annotations from the field if they exist
        field_store = getattr(field, "store", None)
        if field_store is not None:
            if (
                len(field_store.annotate) == 1
                and _annotate_placeholder in field_store.annotate
            ):
                # This is a special case where we need to update the field name,
                # because when field_store was created on __init__,
                # the field name wasn't available.
                # This allows for annotate expressions to be declared as:
                #   total: int = gql.django.field(annotate=Sum("price"))  # noqa: ERA001
                # Instead of the more redundant:
                #   total: int = gql.django.field(annotate={"total": Sum("price")})  # noqa: ERA001
                field_store.annotate = {
                    field.name: field_store.annotate[_annotate_placeholder],
                }
            store |= (
                field_store.with_prefix(prefix, info=info) if prefix else field_store
            )

        # Then from the model property if one is defined
        model_attr = getattr(model, field.python_name, None)
        if model_attr is not None and isinstance(model_attr, ModelProperty):
            attr_store = model_attr.store
            store |= attr_store.with_prefix(prefix, info=info) if prefix else attr_store

        # Lastly, from the django field itself
        model_fieldname: str = getattr(field, "django_name", None) or field.python_name
        model_field = model_fields.get(model_fieldname, None)
        if model_field is not None:
            path = f"{prefix}{model_fieldname}"

            if isinstance(model_field, (models.ForeignKey, OneToOneRel)):
                store.only.append(path)
                store.select_related.append(path)

                # If adding a reverse relation, make sure to select its pointer to us,
                # or else this might causa a refetch from the database
                if isinstance(model_field, OneToOneRel):
                    remote_field = model_field.remote_field
                    store.only.append(
                        f"{path}{LOOKUP_SEP}{resolve_model_field_name(remote_field)}",
                    )

                for f_type_def in get_possible_type_definitions(field.type):
                    f_model = model_field.related_model
                    f_store = _get_model_hints(
                        f_model,
                        schema,
                        f_type_def,
                        f_selection,
                        info=info,
                        config=config,
                        cache=cache,
                        level=level + 1,
                    )
                    if f_store is not None:
                        cache.setdefault(f_model, []).append((level, f_store))
                        store |= f_store.with_prefix(path, info=info)
            elif GenericForeignKey and isinstance(model_field, GenericForeignKey):
                # There's not much we can do to optimize generic foreign keys regarding
                # only/select_related because they can be anything.
                # Just prefetch_related them
                store.prefetch_related.append(model_fieldname)
            elif isinstance(
                model_field,
                _relation_fields,
            ):
                f_types = list(get_possible_type_definitions(field.type))
                if len(f_types) > 1:
                    # This might be a generic foreign key.
                    # In this case, just prefetch it
                    store.prefetch_related.append(model_fieldname)
                elif len(f_types) == 1:
                    remote_field = model_field.remote_field
                    remote_model = remote_field.model
                    f_store = _get_model_hints(
                        remote_model,
                        schema,
                        f_types[0],
                        f_selection,
                        info=info,
                        config=config,
                        cache=cache,
                        level=level + 1,
                    )

                    if f_store is not None:
                        if (
                            (config is None or config.enable_only)
                            and f_store.only
                            and not isinstance(remote_field, ManyToManyRel)
                        ):
                            # If adding a reverse relation, make sure to select its
                            # pointer to us, or else this might causa a refetch from
                            # the database
                            if GenericRelation is not None and isinstance(
                                model_field,
                                GenericRelation,
                            ):
                                f_store.only.append(model_field.object_id_field_name)
                                f_store.only.append(model_field.content_type_field_name)
                            else:
                                f_store.only.append(
                                    remote_field.attname or remote_field.name,
                                )

                        path_lookup = f"{path}{LOOKUP_SEP}"
                        if store.only and f_store.only:
                            extra_only = [
                                o for o in store.only or [] if o.startswith(path_lookup)
                            ]
                            store.only = [o for o in store.only if o not in extra_only]
                            f_store.only.extend(
                                o[len(path_lookup) :] for o in extra_only
                            )

                        if store.select_related and f_store.select_related:
                            extra_sr = [
                                o
                                for o in store.select_related or []
                                if o.startswith(path_lookup)
                            ]
                            store.select_related = [
                                o for o in store.select_related if o not in extra_sr
                            ]
                            f_store.select_related.extend(
                                o[len(path_lookup) :] for o in extra_sr
                            )

                        cache.setdefault(remote_model, []).append((level, f_store))

                        # If prefetch_custom_queryset is false, use _base_manager here
                        # instead of _default_manager because we are getting related
                        # objects, and not querying it directly. Else use the type's
                        # get_queryset and model's custom QuerySet.
                        base_qs = _get_prefetch_queryset(
                            remote_model,
                            field,
                            config,
                            info,
                        )
                        f_qs = f_store.apply(
                            base_qs,
                            info=info,
                            config=config,
                        )
                        f_prefetch = Prefetch(path, queryset=f_qs)
                        f_prefetch._optimizer_sentinel = _sentinel  # type: ignore
                        store.prefetch_related.append(f_prefetch)
            else:
                store.only.append(path)

    # Django keeps track of known fields. That means that if one model select_related or
    # prefetch_related another one, and later another one select_related or
    # prefetch_related the model again, if the used fields there where not optimized in
    # this call django would have to fetch those again. By mergint those with us we are
    # making sure to avoid that
    for inner_level, inner_store in cache.get(model, []):
        if inner_level > level and inner_store:
            # We only want the only/select_related from this. prefetch_related is
            # something else
            store.only.extend(inner_store.only)
            store.select_related.extend(inner_store.select_related)

    return store


def optimize(
    qs: QuerySet[_M] | BaseManager[_M],
    info: GraphQLResolveInfo | Info,
    *,
    config: OptimizerConfig | None = None,
    store: OptimizerStore | None = None,
) -> QuerySet[_M]:
    """Optimize the given queryset considering the gql info.

    This will look through the gql selections, fields and model hints and apply
    `only`, `select_related`, `prefetch_related` and `annotate` optimizations
    according those on the `QuerySet`_.

    Note:
    ----
        This do not execute the queryset, it only optimizes it for when it is actually
        executed.

        It will also avoid doing any extra optimization if the queryset already has
        cached results in it, to avoid triggering extra queries later.

    Args:
    ----
        qs:
            The queryset to be optimized
        info:
            The current field execution info
        config:
            Optional config to use when doing the optimization
        store:
            Optional initial store to use for the optimization

    Returns:
    -------
        The optimized queryset

    .. _QuerySet:
        https://docs.djangoproject.com/en/dev/ref/models/querysets/

    """
    if isinstance(qs, BaseManager):
        qs = cast(QuerySet[_M], qs.all())

    # Avoid optimizing twice and also modify an already resolved queryset
    if (
        getattr(qs, "_gql_optimized", False) or qs._result_cache is not None  # type: ignore
    ):
        return qs

    if isinstance(info, Info):
        info = info._raw_info

    config = config or OptimizerConfig()
    store = store or OptimizerStore()
    schema = cast(Schema, info.schema._strawberry_schema)  # type: ignore

    field_name = info.field_name
    gql_type = get_named_type(info.return_type)
    strawberry_type = schema.get_type_by_name(gql_type.name)
    if strawberry_type is None:
        return qs

    for object_definition in get_possible_type_definitions(strawberry_type):
        if object_definition.is_interface:
            interface_definitions = _interfaces[schema].get(object_definition)
            if interface_definitions is None:
                interface_definitions = []
                for t in schema.schema_converter.type_map.values():
                    t_definition = t.definition
                    if isinstance(
                        t_definition, StrawberryObjectDefinition
                    ) and issubclass(t_definition.origin, object_definition.origin):
                        interface_definitions.append(t_definition)
                _interfaces[schema][object_definition] = interface_definitions

            object_definitions = []
            for interface_definition in interface_definitions:
                dj_definition = get_django_definition(interface_definition.origin)
                if dj_definition and issubclass(qs.model, dj_definition.model):
                    object_definitions.append(interface_definition)
        else:
            object_definitions = [object_definition]

        for selection in convert_selections(info, info.field_nodes):
            if isinstance(selection, InlineFragment) or selection.name != field_name:
                continue

            for t_definition in object_definitions:
                new_store = _get_model_hints(
                    qs.model,
                    schema,
                    t_definition,
                    selection,
                    info=info,
                    config=config,
                )
                if new_store is not None:
                    store |= new_store

    if store:
        qs = store.apply(qs, info=info, config=config)
        qs._gql_optimized = True  # type: ignore

    return qs


optimizer: contextvars.ContextVar[DjangoOptimizerExtension | None] = (
    contextvars.ContextVar(
        "optimizer_ctx",
        default=None,
    )
)


class DjangoOptimizerExtension(SchemaExtension):
    """Automatically optimize returned querysets from internal resolvers.

    Attributes
    ----------
        enable_only_optimization:
            Enable `QuerySet.only` optimizations
        enable_select_related_optimization:
            Enable `QuerySet.select_related` optimizations
        enable_prefetch_related_optimization:
            Enable `QuerySet.prefetch_related` optimizations
        enable_annotate_optimization:
            Enable `QuerySet.annotate` optimizations

    Examples
    --------
        Add the following to your schema configuration.

        >>> import strawberry
        >>> from strawberry_django_plus.optimizer import DjangoOptimizerExtension
        ...
        >>> schema = strawberry.Schema(
        ...     Query,
        ...     extensions=[
        ...         DjangoOptimizerExtension(),
        ...     ]
        ... )

    """

    enabled: contextvars.ContextVar[bool] = contextvars.ContextVar(
        "optimizer_enabled_ctx",
        default=True,
    )

    def __init__(
        self,
        *,
        enable_only_optimization: bool = True,
        enable_select_related_optimization: bool = True,
        enable_prefetch_related_optimization: bool = True,
        enable_annotate_optimization: bool = True,
        execution_context: ExecutionContext | None = None,
        prefetch_custom_queryset: bool = False,
    ):
        super().__init__(execution_context=execution_context)  # type: ignore
        self.enable_only = enable_only_optimization
        self.enable_select_related = enable_select_related_optimization
        self.enable_prefetch_related = enable_prefetch_related_optimization
        self.enable_annotate_optimization = enable_annotate_optimization
        self.prefetch_custom_queryset = prefetch_custom_queryset

    def on_execute(self) -> Generator[None, None, None]:
        token = optimizer.set(self)
        try:
            yield
        finally:
            optimizer.reset(token)

    def resolve(
        self,
        _next: Callable,
        root: Any,
        info: GraphQLResolveInfo,
        *args,
        **kwargs,
    ) -> AwaitableOrValue[Any]:
        ret = _next(root, info, *args, **kwargs)
        if not self.enabled.get():
            return ret

        if isinstance(ret, BaseManager):
            ret = ret.all()

        if isinstance(ret, QuerySet) and ret._result_cache is None:  # type: ignore
            config = OptimizerConfig(
                enable_only=(
                    self.enable_only and info.operation.operation == OperationType.QUERY
                ),
                enable_select_related=self.enable_select_related,
                enable_prefetch_related=self.enable_prefetch_related,
                enable_annotate=self.enable_annotate_optimization,
                prefetch_custom_queryset=self.prefetch_custom_queryset,
            )
            ret = django_fetch(optimize(qs=ret, info=info, config=config))

        return ret

    @classmethod
    @contextlib.contextmanager
    def disabled(cls):
        token = cls.enabled.set(False)
        try:
            yield
        finally:
            cls.enabled.reset(token)

    def optimize(
        self,
        qs: QuerySet[_M] | BaseManager[_M],
        info: GraphQLResolveInfo | Info,
        *,
        store: OptimizerStore | None = None,
    ) -> QuerySet[_M]:
        if not self.enabled.get():
            return qs

        config = OptimizerConfig(
            enable_only=self.enable_only
            and info.operation.operation == OperationType.QUERY,
            enable_select_related=self.enable_select_related,
            enable_prefetch_related=self.enable_prefetch_related,
            enable_annotate=self.enable_annotate_optimization,
            prefetch_custom_queryset=self.prefetch_custom_queryset,
        )
        return optimize(qs, info, config=config, store=store)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Iterable, Union

import strawberry
from django.core.exceptions import (
    NON_FIELD_ERRORS,
    ObjectDoesNotExist,
    PermissionDenied,
    ValidationError,
)
from django.db import models, transaction
from strawberry import UNSET, relay
from strawberry.annotation import StrawberryAnnotation
from strawberry.field import UNRESOLVED
from strawberry.utils.str_converters import capitalize_first, to_camel_case
from typing_extensions import Annotated

from strawberry_django.arguments import argument
from strawberry_django.fields.field import (
    StrawberryDjangoFieldBase,
    StrawberryDjangoFieldFilters,
)
from strawberry_django.fields.types import OperationInfo, OperationMessage
from strawberry_django.optimizer import DjangoOptimizerExtension
from strawberry_django.permissions import filter_with_perms, get_with_perms
from strawberry_django.resolvers import django_resolver
from strawberry_django.settings import strawberry_django_settings
from strawberry_django.utils.inspect import get_possible_types

from . import resolvers

if TYPE_CHECKING:
    from graphql.pyutils import AwaitableOrValue
    from strawberry.arguments import StrawberryArgument
    from strawberry.type import StrawberryType, WithStrawberryObjectDefinition
    from strawberry.types import Info
    from strawberry.types.types import StrawberryObjectDefinition
    from typing_extensions import Literal, Self


def _get_validation_errors(error: Exception):
    if isinstance(error, PermissionDenied):
        kind = OperationMessage.Kind.PERMISSION
    elif isinstance(error, ValidationError):
        kind = OperationMessage.Kind.VALIDATION
    elif isinstance(error, ObjectDoesNotExist):
        kind = OperationMessage.Kind.ERROR
    else:
        kind = OperationMessage.Kind.ERROR

    if isinstance(error, ValidationError) and hasattr(error, "error_dict"):
        # convert field errors
        for field, field_errors in error.error_dict.items():
            for e in field_errors:
                yield OperationMessage(
                    kind=kind,
                    field=to_camel_case(field) if field != NON_FIELD_ERRORS else None,
                    message=e.message % e.params if e.params else e.message,
                    code=getattr(e, "code", None),
                )
    elif isinstance(error, ValidationError) and hasattr(error, "error_list"):
        # convert non-field errors
        for e in error.error_list:
            yield OperationMessage(
                kind=kind,
                message=e.message % e.params if e.params else e.message,
                code=getattr(error, "code", None),
            )
    else:
        msg = getattr(error, "msg", None)
        if msg is None:
            msg = str(error)

        yield OperationMessage(
            kind=kind,
            message=msg,
            code=getattr(error, "code", None),
        )


def _handle_exception(error: Exception):
    if isinstance(error, (ValidationError, PermissionDenied, ObjectDoesNotExist)):
        return OperationInfo(
            messages=list(_get_validation_errors(error)),
        )

    raise error


class DjangoMutationBase(StrawberryDjangoFieldBase):
    def __init__(
        self,
        *args,
        handle_django_errors: bool | None = None,
        **kwargs,
    ):
        self._resolved_return_type: bool = False

        if handle_django_errors is None:
            settings = strawberry_django_settings()
            handle_django_errors = settings["MUTATIONS_DEFAULT_HANDLE_ERRORS"]
        self.handle_errors = handle_django_errors

        super().__init__(*args, **kwargs)

    def __copy__(self) -> Self:
        new_field = super().__copy__()
        new_field.handle_errors = self.handle_errors
        return new_field

    def resolve_type(
        self,
        *,
        type_definition: StrawberryObjectDefinition | None = None,
    ) -> (
        StrawberryType | type[WithStrawberryObjectDefinition] | Literal[UNRESOLVED]  # type: ignore
    ):
        resolved = super().resolve_type(type_definition=type_definition)
        if resolved is UNRESOLVED:
            return resolved

        if self.handle_errors and not self._resolved_return_type:
            types_ = tuple(get_possible_types(resolved))
            if OperationInfo not in types_:
                types_ = (*types_, OperationInfo)

                name = capitalize_first(to_camel_case(self.python_name))
                resolved = Annotated[
                    Union[types_],  # type: ignore
                    strawberry.union(f"{name}Payload"),
                ]
                self.type_annotation = StrawberryAnnotation(
                    resolved,
                    namespace=getattr(self.type_annotation, "namespace", None),
                )

            self._resolved_return_type = True

        return resolved

    def get_result(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> AwaitableOrValue[Any]:
        if not self.handle_errors:
            return self.resolver(source, info, args, kwargs)

        # TODO: Any other exception types that we should capture here?
        try:
            resolved = self.resolver(source, info, args, kwargs)
        except Exception as e:  # noqa: BLE001
            return _handle_exception(e)

        if inspect.isawaitable(resolved):

            async def async_resolver():
                try:
                    return await resolved
                except Exception as e:  # noqa: BLE001
                    return _handle_exception(e)

            return async_resolver()

        return resolved


class DjangoMutationCUD(DjangoMutationBase):
    def __init__(
        self,
        input_type: type | None = None,
        full_clean: bool = True,
        argument_name: str | None = None,
        key_attr: str | None = "pk",
        **kwargs,
    ):
        self.full_clean = full_clean
        self.input_type = input_type
        self.key_attr = key_attr

        if argument_name is None:
            settings = strawberry_django_settings()
            argument_name = settings["MUTATIONS_DEFAULT_ARGUMENT_NAME"]
        self.argument_name = argument_name

        super().__init__(**kwargs)

    def __copy__(self) -> Self:
        new_field = super().__copy__()
        new_field.input_type = self.input_type
        new_field.full_clean = self.full_clean
        return new_field

    @property
    def arguments(self):
        arguments = super().arguments
        if not self.input_type:
            return arguments

        return [
            *arguments,
            argument(
                self.argument_name,
                self.input_type,
                is_list=self.is_list and isinstance(self, DjangoCreateMutation),
            ),
        ]

    @arguments.setter
    def arguments(self, value: list[StrawberryArgument]):
        args_prop = super(DjangoMutationBase, self.__class__).arguments
        return args_prop.fset(self, value)  # type: ignore


class DjangoCreateMutation(DjangoMutationCUD, StrawberryDjangoFieldFilters):
    @django_resolver
    @transaction.atomic
    def resolver(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> Any:
        assert info is not None

        data: list[Any] | Any = kwargs.get(self.argument_name)

        if self.is_list:
            assert isinstance(data, list)
            return [
                self.create(
                    resolvers.parse_input(info, vars(d), key_attr=self.key_attr),
                    info=info,
                )
                for d in data
            ]

        assert not isinstance(data, list)
        return self.create(
            resolvers.parse_input(info, vars(data), key_attr=self.key_attr)
            if data is not None
            else {},
            info=info,
        )

    def create(self, data: dict[str, Any], *, info: Info):
        model = self.django_model
        assert model is not None

        # Do not optimize anything while retrieving the object to create
        with DjangoOptimizerExtension.disabled():
            return resolvers.create(
                info,
                model,
                data,
                key_attr=self.key_attr,
                full_clean=self.full_clean,
            )


def get_pk(
    data: dict[str, Any],
    *,
    key_attr: str | None = "pk",
) -> strawberry.ID | relay.GlobalID | Literal[UNSET] | None:  # type: ignore
    pk = data.pop(key_attr, UNSET) if key_attr else UNSET

    if pk is UNSET:
        pk = data.pop("id", UNSET)
    return pk


class DjangoUpdateMutation(DjangoMutationCUD, StrawberryDjangoFieldFilters):
    @django_resolver
    @transaction.atomic
    def resolver(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> Any:
        assert info is not None

        model = self.django_model
        assert model is not None

        data: Any = kwargs.get(self.argument_name)
        vdata = vars(data).copy() if data is not None else {}

        pk = get_pk(vdata, key_attr=self.key_attr)
        if pk not in (None, UNSET):  # noqa: PLR6201
            instance = get_with_perms(
                pk,
                info,
                required=True,
                model=model,
                key_attr=self.key_attr,
            )
        else:
            instance = filter_with_perms(
                self.get_queryset(
                    queryset=model._default_manager.all(),
                    info=info,
                    **kwargs,
                ),
                info,
            )

        return self.update(
            info, instance, resolvers.parse_input(info, vdata, key_attr=self.key_attr)
        )

    def update(
        self,
        info: Info,
        instance: models.Model | Iterable[models.Model],
        data: dict[str, Any],
    ):
        # Do not optimize anything while retrieving the object to update
        with DjangoOptimizerExtension.disabled():
            return resolvers.update(
                info,
                instance,
                data,
                key_attr=self.key_attr,
                full_clean=self.full_clean,
            )


class DjangoDeleteMutation(
    DjangoMutationCUD,
    DjangoMutationBase,
    StrawberryDjangoFieldFilters,
):
    @django_resolver
    @transaction.atomic
    def resolver(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> Any:
        assert info is not None

        model = self.django_model
        assert model is not None

        data: Any = kwargs.get(self.argument_name)
        vdata = vars(data).copy() if data is not None else {}

        pk = get_pk(vdata, key_attr=self.key_attr)
        if pk not in (None, UNSET):  # noqa: PLR6201
            instance = get_with_perms(
                pk,
                info,
                required=True,
                model=model,
                key_attr=self.key_attr,
            )
        else:
            instance = filter_with_perms(
                self.get_queryset(
                    queryset=model._default_manager.all(),
                    info=info,
                    **kwargs,
                ),
                info,
            )

        return self.delete(
            info, instance, resolvers.parse_input(info, vdata, key_attr=self.key_attr)
        )

    def delete(
        self,
        info: Info,
        instance: models.Model | Iterable[models.Model],
        data: dict[str, Any] | None = None,
    ):
        # Do not optimize anything while retrieving the object to update
        with DjangoOptimizerExtension.disabled():
            return resolvers.delete(
                info,
                instance,
                data=data,
            )



from __future__ import annotations

import dataclasses
import inspect
from functools import cached_property
from typing import TYPE_CHECKING, Any, Final, Literal, Optional, Sequence, overload

from strawberry import UNSET
from strawberry.annotation import StrawberryAnnotation
from strawberry.exceptions import MissingArgumentsAnnotationsError
from strawberry.field import StrawberryField
from strawberry.types.fields.resolver import ReservedName, StrawberryResolver
from typing_extensions import Self

from strawberry_django.exceptions import (
    ForbiddenFieldArgumentError,
    MissingFieldArgumentError,
)
from strawberry_django.utils.typing import is_auto

if TYPE_CHECKING:
    from collections.abc import Callable, MutableMapping

    from strawberry.extensions.field_extension import FieldExtension
    from strawberry.field import _RESOLVER_TYPE, T
    from strawberry.types import Info


QUERYSET_PARAMSPEC = ReservedName("queryset")
PREFIX_PARAMSPEC = ReservedName("prefix")
SEQUENCE_PARAMSPEC = ReservedName("sequence")
VALUE_PARAM = ReservedName("value")

OBJECT_FILTER_NAME: Final[str] = "filter"
OBJECT_ORDER_NAME: Final[str] = "order"
WITH_NONE_META: Final[str] = "WITH_NONE_META"
RESOLVE_VALUE_META: Final[str] = "RESOLVE_VALUE_META"


class FilterOrderFieldResolver(StrawberryResolver):
    RESERVED_PARAMSPEC = (
        *StrawberryResolver.RESERVED_PARAMSPEC,
        QUERYSET_PARAMSPEC,
        PREFIX_PARAMSPEC,
        SEQUENCE_PARAMSPEC,
        VALUE_PARAM,
    )

    def __init__(self, *args, resolver_type: Literal["filter", "order"], **kwargs):
        super().__init__(*args, **kwargs)
        self._resolver_type = resolver_type

    def validate_filter_arguments(self):
        is_object_filter = self.name == OBJECT_FILTER_NAME
        is_object_order = self.name == OBJECT_ORDER_NAME

        if not self.reserved_parameters[PREFIX_PARAMSPEC]:
            raise MissingFieldArgumentError(PREFIX_PARAMSPEC.name, self)

        if (is_object_filter or is_object_order) and not self.reserved_parameters[
            QUERYSET_PARAMSPEC
        ]:
            raise MissingFieldArgumentError(QUERYSET_PARAMSPEC.name, self)

        if (
            self._resolver_type != OBJECT_ORDER_NAME
            and self.reserved_parameters[SEQUENCE_PARAMSPEC]
        ):
            raise ForbiddenFieldArgumentError(self, [SEQUENCE_PARAMSPEC.name])

        value_param = self.reserved_parameters[VALUE_PARAM]
        if value_param:
            if is_object_filter or is_object_order:
                raise ForbiddenFieldArgumentError(self, [VALUE_PARAM.name])

            annotation = self.strawberry_annotations[value_param]
            if annotation is None:
                raise MissingArgumentsAnnotationsError(self, [VALUE_PARAM.name])
        elif not is_object_filter and not is_object_order:
            raise MissingFieldArgumentError(VALUE_PARAM.name, self)

        parameters = self.signature.parameters.values()
        reserved_parameters = set(self.reserved_parameters.values())
        exta_params = [p for p in parameters if p not in reserved_parameters]
        if exta_params:
            raise ForbiddenFieldArgumentError(self, [p.name for p in exta_params])

    @cached_property
    def type_annotation(self) -> StrawberryAnnotation | None:
        param = self.reserved_parameters[VALUE_PARAM]
        if param and param is not inspect.Signature.empty:
            annotation = param.annotation
            if is_auto(annotation) and self._resolver_type == OBJECT_ORDER_NAME:
                from strawberry_django import ordering

                annotation = ordering.Ordering

            return StrawberryAnnotation(Optional[annotation])

        return None

    def __call__(  # type: ignore
        self,
        source: Any,
        info: Info | None,
        queryset=None,
        sequence=None,
        **kwargs: Any,
    ) -> Any:
        args = []

        if self.self_parameter:
            args.append(source)

        if parent_parameter := self.parent_parameter:
            kwargs[parent_parameter.name] = source

        if root_parameter := self.root_parameter:
            kwargs[root_parameter.name] = source

        if info_parameter := self.info_parameter:
            assert info is not None
            kwargs[info_parameter.name] = info

        if info_parameter := self.reserved_parameters.get(QUERYSET_PARAMSPEC):
            assert queryset is not None
            kwargs[info_parameter.name] = queryset

        if info_parameter := self.reserved_parameters.get(SEQUENCE_PARAMSPEC):
            assert sequence is not None
            kwargs[info_parameter.name] = sequence

        return super().__call__(*args, **kwargs)


class FilterOrderField(StrawberryField):
    base_resolver: FilterOrderFieldResolver | None  # type: ignore

    def __call__(self, resolver: _RESOLVER_TYPE) -> Self | FilterOrderFieldResolver:  # type: ignore
        if not isinstance(resolver, StrawberryResolver):
            resolver = FilterOrderFieldResolver(
                resolver, resolver_type=self.metadata["_FIELD_TYPE"]
            )
        elif not isinstance(resolver, FilterOrderFieldResolver):
            raise TypeError(
                'Expected resolver to be instance of "FilterOrderFieldResolver", '
                f'found "{type(resolver)}"'
            )

        super().__call__(resolver)
        self._arguments = []
        resolver.validate_filter_arguments()

        if resolver.name in {OBJECT_FILTER_NAME, OBJECT_ORDER_NAME}:
            # For object filter we return resolver
            return resolver

        self.init = self.compare = self.repr = True
        return self


@overload
def filter_field(
    *,
    resolver: _RESOLVER_TYPE[T],
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    init: Literal[False] = False,
    deprecation_reason: str | None = None,
    default: Any = UNSET,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: MutableMapping[Any, Any] | None = None,
    directives: Sequence[object] = (),
    extensions: list[FieldExtension] | None = None,
    filter_none: bool = False,
    resolve_value: bool = UNSET,
) -> T: ...


@overload
def filter_field(
    *,
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    init: Literal[True] = True,
    deprecation_reason: str | None = None,
    default: Any = UNSET,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: MutableMapping[Any, Any] | None = None,
    directives: Sequence[object] = (),
    extensions: list[FieldExtension] | None = None,
    filter_none: bool = False,
    resolve_value: bool = UNSET,
) -> Any: ...


@overload
def filter_field(
    resolver: _RESOLVER_TYPE[T],
    *,
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    deprecation_reason: str | None = None,
    default: Any = UNSET,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: MutableMapping[Any, Any] | None = None,
    directives: Sequence[object] = (),
    extensions: list[FieldExtension] | None = None,
    filter_none: bool = False,
    resolve_value: bool = UNSET,
) -> StrawberryField: ...


def filter_field(
    resolver: _RESOLVER_TYPE[Any] | None = None,
    *,
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    deprecation_reason: str | None = None,
    default: Any = UNSET,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: MutableMapping[Any, Any] | None = None,
    directives: Sequence[object] = (),
    extensions: list[FieldExtension] | None = None,
    filter_none: bool = False,
    resolve_value: bool = UNSET,
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> Any:
    """Annotates a method or property as a Django filter field.

    If using with method, these parameters are required: queryset, value, prefix
    Additionaly value has to be annotated with type of filter

    This is normally used inside a type declaration:

    >>> @strawberry_django.filter(SomeModel)
    >>> class X:
    >>>     field_abc: strawberry.auto = strawberry_django.filter_field()

    >>>     @strawberry.filter_field(description="ABC")
    >>>     def field_with_resolver(self, queryset, info, value: str, prefix):
    >>>         return

    it can be used both as decorator and as a normal function.
    """
    metadata = metadata or {}
    metadata["_FIELD_TYPE"] = OBJECT_FILTER_NAME
    metadata[RESOLVE_VALUE_META] = resolve_value
    metadata[WITH_NONE_META] = filter_none

    field_ = FilterOrderField(
        python_name=None,
        graphql_name=name,
        is_subscription=is_subscription,
        description=description,
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        extensions=extensions or [],
    )

    if resolver:
        return field_(resolver)

    return field_


@overload
def order_field(
    *,
    resolver: _RESOLVER_TYPE[T],
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    init: Literal[False] = False,
    deprecation_reason: str | None = None,
    default: Any = UNSET,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: MutableMapping[Any, Any] | None = None,
    directives: Sequence[object] = (),
    extensions: list[FieldExtension] | None = None,
    order_none: bool = False,
) -> T: ...


@overload
def order_field(
    *,
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    init: Literal[True] = True,
    deprecation_reason: str | None = None,
    default: Any = UNSET,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: MutableMapping[Any, Any] | None = None,
    directives: Sequence[object] = (),
    extensions: list[FieldExtension] | None = None,
    order_none: bool = False,
) -> Any: ...


@overload
def order_field(
    resolver: _RESOLVER_TYPE[T],
    *,
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    deprecation_reason: str | None = None,
    default: Any = UNSET,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: MutableMapping[Any, Any] | None = None,
    directives: Sequence[object] = (),
    extensions: list[FieldExtension] | None = None,
    order_none: bool = False,
) -> StrawberryField: ...


def order_field(
    resolver: _RESOLVER_TYPE[Any] | None = None,
    *,
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    deprecation_reason: str | None = None,
    default: Any = UNSET,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: MutableMapping[Any, Any] | None = None,
    directives: Sequence[object] = (),
    extensions: list[FieldExtension] | None = None,
    order_none: bool = False,
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> Any:
    """Annotates a method or property as a Django filter field.

    If using with method, these parameters are required: queryset, value, prefix
    Additionaly value has to be annotated with type of filter

    This is normally used inside a type declaration:

    >>> @strawberry_django.order(SomeModel)
    >>> class X:
    >>>     field_abc: strawberry.auto = strawberry_django.order_field()

    >>>     @strawberry.order_field(description="ABC")
    >>>     def field_with_resolver(self, queryset, info, value: str, prefix):
    >>>         return

    it can be used both as decorator and as a normal function.
    """
    metadata = metadata or {}
    metadata["_FIELD_TYPE"] = OBJECT_ORDER_NAME
    metadata[WITH_NONE_META] = order_none

    field_ = FilterOrderField(
        python_name=None,
        graphql_name=name,
        is_subscription=is_subscription,
        description=description,
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        extensions=extensions or [],
    )

    if resolver:
        return field_(resolver)

    return field_



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, Optional, TypeVar

from django.db.models import ForeignKey
from strawberry import LazyType, relay
from strawberry.annotation import StrawberryAnnotation
from strawberry.auto import StrawberryAuto
from strawberry.field import UNRESOLVED, StrawberryField
from strawberry.type import (
    StrawberryContainer,
    StrawberryList,
    StrawberryOptional,
    StrawberryType,
    WithStrawberryObjectDefinition,
    get_object_definition,
)
from strawberry.union import StrawberryUnion

from strawberry_django.resolvers import django_resolver
from strawberry_django.utils.typing import (
    WithStrawberryDjangoObjectDefinition,
    get_django_definition,
    has_django_definition,
    unwrap_type,
)

if TYPE_CHECKING:
    from django.db import models
    from strawberry.object_type import StrawberryObjectDefinition
    from strawberry.types import Info
    from typing_extensions import Literal, Self

    from strawberry_django.type import StrawberryDjangoDefinition

_QS = TypeVar("_QS", bound="models.QuerySet")


class StrawberryDjangoFieldBase(StrawberryField):
    def __init__(
        self,
        django_name: str | None = None,
        graphql_name: str | None = None,
        python_name: str | None = None,
        **kwargs,
    ):
        self.is_relation = False
        self.django_name = django_name
        self.origin_django_type: StrawberryDjangoDefinition[Any, Any] | None = None
        super().__init__(graphql_name=graphql_name, python_name=python_name, **kwargs)

    def __copy__(self) -> Self:
        new_field = super().__copy__()
        new_field.django_name = self.django_name
        new_field.is_relation = self.is_relation
        new_field.origin_django_type = self.origin_django_type
        return new_field

    @property
    def is_basic_field(self) -> bool:
        """Mark this field as not basic.

        All StrawberryDjango fields define a custom resolver that needs to be
        run, so always return False here.
        """
        return False

    @functools.cached_property
    def is_async(self) -> bool:
        # Our default resolver is sync by default but will return a coroutine
        # when running ASGI. If we happen to have an extension that only supports
        # async, make sure we mark the field as async as well to support resolving
        # it properly.
        return super().is_async or any(
            e.supports_async and not e.supports_sync for e in self.extensions
        )

    @functools.cached_property
    def django_type(self) -> type[WithStrawberryDjangoObjectDefinition] | None:
        origin = self.type

        object_definition = get_object_definition(origin)

        if object_definition and issubclass(object_definition.origin, relay.Connection):
            origin = object_definition.type_var_map.get("NodeType")

            if origin is None:
                specialized_type_var_map = (
                    object_definition.specialized_type_var_map or {}
                )

                origin = specialized_type_var_map["NodeType"]

            if isinstance(origin, LazyType):
                origin = origin.resolve_type()

        origin = unwrap_type(origin)
        if isinstance(origin, LazyType):
            origin = origin.resolve_type()

        if isinstance(origin, StrawberryUnion):
            origin_list: list[type[WithStrawberryDjangoObjectDefinition]] = []
            for t in origin.types:
                while isinstance(t, StrawberryContainer):
                    t = t.of_type  # noqa: PLW2901

                if has_django_definition(t):
                    origin_list.append(t)

            origin = origin_list[0] if len(origin_list) == 1 else None

        return origin if has_django_definition(origin) else None

    @functools.cached_property
    def django_model(self) -> type[models.Model] | None:
        django_type = self.django_type
        return (
            django_type.__strawberry_django_definition__.model
            if django_type is not None
            else None
        )

    @functools.cached_property
    def is_optional(self) -> bool:
        return isinstance(self.type, StrawberryOptional)

    @functools.cached_property
    def is_list(self) -> bool:
        type_ = self.type
        if isinstance(type_, StrawberryOptional):
            type_ = type_.of_type

        return isinstance(type_, StrawberryList)

    @functools.cached_property
    def is_connection(self) -> bool:
        type_ = self.type
        if isinstance(type_, StrawberryOptional):
            type_ = type_.of_type

        return isinstance(type_, type) and issubclass(type_, relay.Connection)

    @functools.cached_property
    def safe_resolver(self):
        resolver = self.base_resolver
        assert resolver

        if not resolver.is_async:
            resolver = django_resolver(resolver, qs_hook=None)

        return resolver

    def resolve_type(
        self,
        *,
        type_definition: StrawberryObjectDefinition | None = None,
    ) -> (
        StrawberryType | type[WithStrawberryObjectDefinition] | Literal[UNRESOLVED]  # type: ignore
    ):
        resolved = super().resolve_type(type_definition=type_definition)
        if resolved is UNRESOLVED:
            return resolved

        resolved_django_type = get_django_definition(unwrap_type(resolved))

        if self.origin_django_type and (
            # FIXME: Why does this come as Any sometimes when using future annotations?
            resolved is Any
            or isinstance(resolved, StrawberryAuto)
            # If the resolved type is an input but the origin is not, or vice versa,
            # resolve this again
            or (
                resolved_django_type
                and resolved_django_type.is_input != self.origin_django_type.is_input
            )
        ):
            from .types import get_model_field, is_optional, resolve_model_field_type

            model_field = get_model_field(
                self.origin_django_type.model,
                self.django_name or self.python_name or self.name,
            )
            resolved_type = resolve_model_field_type(
                (
                    model_field.target_field
                    if (
                        self.python_name.endswith("_id")
                        and isinstance(model_field, ForeignKey)
                    )
                    else model_field
                ),
                self.origin_django_type,
            )
            if is_optional(
                model_field,
                self.origin_django_type.is_input,
                self.origin_django_type.is_partial,
            ):
                resolved_type = Optional[resolved_type]

            self.type_annotation = StrawberryAnnotation(resolved_type)
            resolved = super().type

        if isinstance(resolved, StrawberryAuto):
            resolved = UNRESOLVED

        return resolved

    def resolver(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> Any:
        return self.safe_resolver(*args, **kwargs)

    def get_result(self, source, info, args, kwargs):
        return self.resolver(info, source, args, kwargs)

    def get_queryset(self, queryset: _QS, info: Info, **kwargs) -> _QS:
        return queryset



from functools import _make_key  # noqa: PLC2701
from typing import Callable, Dict, Hashable, Optional, Tuple, cast

from django.core.cache import caches
from django.core.cache.backends.base import DEFAULT_TIMEOUT
from strawberry.extensions import SchemaExtension
from strawberry.types import ExecutionContext


class DjangoCacheBase(SchemaExtension):
    """Base for a Cache that uses Django built in cache instead of an in memory cache.

    Arguments:
    ---------
    `cache_name: str`
        Name of the Django Cache to use, defaults to 'default'

    `timeout: Optional[int]`
        How long to hold items in the cache. See the Django Cache docs for details
        https://docs.djangoproject.com/en/4.0/topics/cache/

    `hash_fn: Optional[Callable[[Tuple, Dict], str]]`
        A function to use to generate the cache keys
        Defaults to the same key generator as functools.lru_cache
        WARNING! The default function does NOT work with memcached
        and will generate warnings

    """

    def __init__(
        self,
        cache_name: str = "default",
        timeout: Optional[int] = None,
        hash_fn: Optional[Callable[[Tuple, Dict], Hashable]] = None,
        *,
        execution_context: Optional[ExecutionContext] = None,
    ):
        super().__init__(execution_context=cast(ExecutionContext, execution_context))

        self.cache = caches[cache_name]
        self.timeout = timeout or DEFAULT_TIMEOUT
        # Use same key generating function as functools.lru_cache as default
        self.hash_fn = hash_fn or (lambda args, kwargs: _make_key(args, kwargs, False))

    def execute_cached(self, func, *args, **kwargs):
        hash_key = cast(str, self.hash_fn(args, kwargs))
        cache_result = self.cache.get(hash_key)
        if cache_result is not None:
            return cache_result

        func_result = func(*args, **kwargs)
        self.cache.set(hash_key, func_result, timeout=self.timeout)
        return func_result



from typing import Any, Dict, Mapping, TypeVar

from typing_extensions import TypeAlias

_K = TypeVar("_K", bound=Any)
_V = TypeVar("_V", bound=Any)

DictTree: TypeAlias = Dict[str, "DictTree"]


def dicttree_merge(dict1: Mapping[_K, _V], dict2: Mapping[_K, _V]) -> Dict[_K, _V]:
    new = {
        **dict1,
        **dict2,
    }

    for k, v1 in dict1.items():
        if not isinstance(v1, dict):
            continue

        v2 = dict2.get(k)
        if isinstance(v2, Mapping):
            new[k] = dicttree_merge(v1, v2)  # type: ignore

    for k, v2 in dict2.items():
        if not isinstance(v2, dict):
            continue

        v1 = dict1.get(k)
        if isinstance(v1, Mapping):
            new[k] = dicttree_merge(v1, v2)  # type: ignore

    return new


def dicttree_insersection_differs(
    dict1: Mapping[_K, _V],
    dict2: Mapping[_K, _V],
) -> bool:
    for k in set(dict1) & set(dict2):
        v1 = dict1[k]
        v2 = dict2[k]

        if isinstance(v1, Mapping) and isinstance(v2, Mapping):
            if dicttree_insersection_differs(v1, v2):
                return True
        elif v1 != v2:
            return True

    return False
